package Image::Mosaic;
#
# Routines for building photo mosaics - Jim Bumgardner
#
# Works with an Image::Set module, used for delivering a sequence of images.
#
# Currently only Image::FlickrSet is implemented

use Image::Magick;
use Clone qw(clone);

require Exporter;
our @ISA = qw(Exporter);
@EXPORT = qw(sample_photos  make_heatmap  output_mosaic);
@EXPORT_OK = qw(setup_cells  select_tiles);

use strict;
use warnings;

my %fields = (
	max_images => 800,
  reso => 7,
  cellsize => 20,
  noborders => 0,
  verbose => 0,
  grabThumbs => 0,
  doflops => 0,
  tileblur => 0.4,
	tilefilter => 'Sinc',
  targetblur => 0.4,
	targetfilter => 'Sinc',
);

sub new 
{
  my ($class,$options) = @_;
  my $self = {%fields};
  foreach my $key (keys %{$options})
  {
  	$self->{$key} = $options->{$key};
  }
  $self->{reso2} = $self->{reso}**2;
  bless $self, $class;
  return $self;
}

sub CumDiff()
{
  my ($self,$img, $cell, $upperBound) = @_;
  
  my $sum = 0;
  
	my ($pix1,$pix2) = ($cell->{pix}, $img->{pix});
	for (my $i = 0; $i < $self->{reso2}; ++$i)
	{
		my ($p1, $p2) = ($pix1->[$i], $pix2->[$i]);
		$sum += ($p1->[0] - $p2->[0])**2 +
		        ($p1->[1] - $p2->[1])**2 +
		        ($p1->[2] - $p2->[2])**2;
		last if $sum > $upperBound;       
	}

  return $sum;
}

sub CumDiffFlop($$)
{
  my ($self,$img, $cell, $upperBound) = @_;
  
  my $sum = 0;
  
	my ($pix1,$pix2) = ($cell->{pix}, $img->{pix});

  my $r = $self->{reso};
	for (my $i = 0; $i < $self->{reso2}; ++$i)
	{
		my ($p1,$p2) = ($pix1->[$i], $pix2->[int($i/$r)*$r+($r-1)-($i % $r)]);
		$sum += ($p1->[0] - $p2->[0])**2 +
	          ($p1->[1] - $p2->[1])**2 +
	          ($p1->[2] - $p2->[2])**2;
		last if $sum > $upperBound;       
	}
  return $sum;
}

# higher quality edge detection, but slower
sub Edginess()
{
  my ($self,$cell) = @_;

  my $pix = $cell->{pix};

  my $reso21 = $self->{reso2}-1;

  # compute deviation, and raise score for tied cells in center...
  my $cumdiff = 0;

  for (my $i = 0; $i <  $self->{reso2}; ++$i)
  {
    for (my $j = $i+1; $j < $self->{reso2}; ++$j)
    {
    	my $d = ($pix->[$j]->[0] - $pix->[$i]->[0])**2 +
    	        ($pix->[$j]->[1] - $pix->[$i]->[1])**2 +
    	        ($pix->[$j]->[2] - $pix->[$i]->[2])**2;
    	$cumdiff += $d;
    }
  }
  return $cumdiff;
}

# corner-based edge detection...
sub EdginessFast()
{
  my ($self,$cell) = @_;

  my $pix = $cell->{pix};

  my $reso21 = $self->{reso2}-1;

  # compute deviation, and raise score for tied cells in center...
  my $cumdiff = 0;

  my ($tl,$tr,$bl,$br) = (0,$self->{reso}-1,($self->{reso}-1)*$self->{reso},$self->{reso}*$self->{reso} - 1);
  my ($i,$j) = ($tl, $tr);
  $cumdiff += ($pix->[$j]->[0] - $pix->[$i]->[0])**2 +
    	        ($pix->[$j]->[1] - $pix->[$i]->[1])**2 +
    	        ($pix->[$j]->[2] - $pix->[$i]->[2])**2;

  ($i,$j) = ($tl, $bl);
  $cumdiff += ($pix->[$j]->[0] - $pix->[$i]->[0])**2 +
    	        ($pix->[$j]->[1] - $pix->[$i]->[1])**2 +
    	        ($pix->[$j]->[2] - $pix->[$i]->[2])**2;


  ($i,$j) = ($tl, $br);
  $cumdiff += ($pix->[$j]->[0] - $pix->[$i]->[0])**2 +
    	        ($pix->[$j]->[1] - $pix->[$i]->[1])**2 +
    	        ($pix->[$j]->[2] - $pix->[$i]->[2])**2;

  ($i,$j) = ($tr, $bl );
  $cumdiff += ($pix->[$j]->[0] - $pix->[$i]->[0])**2 +
    	        ($pix->[$j]->[1] - $pix->[$i]->[1])**2 +
    	        ($pix->[$j]->[2] - $pix->[$i]->[2])**2;

  ($i,$j) = ($tr, $br );
  $cumdiff += ($pix->[$j]->[0] - $pix->[$i]->[0])**2 +
    	        ($pix->[$j]->[1] - $pix->[$i]->[1])**2 +
    	        ($pix->[$j]->[2] - $pix->[$i]->[2])**2;

  ($i,$j) = ($bl, $br );
  $cumdiff += ($pix->[$j]->[0] - $pix->[$i]->[0])**2 +
    	        ($pix->[$j]->[1] - $pix->[$i]->[1])**2 +
    	        ($pix->[$j]->[2] - $pix->[$i]->[2])**2;

  return $cumdiff;
}


sub RGBtoHSV($$$)
{
  my ($r,$g,$b) = @_;
  my $max = $r > $g? $r : $g;
  $max = $max > $b? $max : $b;
  my $min = $r < $g? $r : $g;
  $min = $min < $b? $min : $b;
	my $v = $max;
	my $s = ($max != 0)? ($max-$min)/$max : 0;
	my $h;
	if ($s == 0) {
		$h = 0;	# undefined, actually
	}
	else {
		my $d = $max - $min;
		if ($r == $max) {
			$h = ($g - $b)/$d;
		}
		elsif ($g == $max) {
			$h = 2 + ($b-$r)/$d;
		}
		elsif ($b == $max) {
			$h = 4 + ($r-$g)/$d;
		}
		$h *= 60;
		if ($h < 0) {
			$h += 360;
		}
	}
	return ($h,$s,$v);
}

sub subsample_photo($)
{
  my ($self,$photo) = @_;
	if (!$photo->{pix})
	{
		my $image = $self->{imageset}->get_image($photo->{idx},$self->{reso});
		warn "Problem getting image $photo->{idx}" if !$image;
		# crop to square
	  my ($w, $h) = $image->Get('width', 'height');
	  if ($w < $h) {
	  	$image->Crop(width=>$w, height=>$w, x=>0, y=>($h-$w)/2);
	  }
	  elsif ($w > $h) {
	  	$image->Crop(width=>$h, height=>$h, y=>0, x=>($w-$h)/2);
	  }
		my $err = $image->Resize(width=>$self->{reso}, height=>$self->{reso}, filter=>$self->{tilefilter}, blur=>$self->{tileblur});
		my @pix = ();
		for (my $py = 0; $py < $self->{reso}; ++$py)
		{
			for (my $px = 0; $px < $self->{reso}; ++$px)
			{
					my ($r,$g,$b) = $image->GetPixels(x=>$px, y=>$py,width=>1,height=>1,normalize=>1);
					push @pix, [$r,$g,$b];
			}
		}

		$photo->{pix} = \@pix;
		undef $image;
	}
}

sub make_heatmap($$)
{
  my ($self, $oname) = @_;
  
  $self->setup_cells() if !$self->{sortedcells};
  return if !$self->{sortedcells};

  my ($width, $height) = ($self->{reso} * $self->{hcells}, $self->{reso} * $self->{vcells});
	
	my $heatmap = Image::Magick->new;
	$heatmap->Set(size=>"${width}x${height}"); # worked okay at 10000x10000
  $heatmap->Read('xc:black');

	print "Heatmap size: $width x $height\n";

  # for each cell
	my $n = 0;  
  for my $cell (@{$self->{sortedcells}})
  {
    my $alpha = (1 - $n / (scalar(@{$self->{sortedcells}}) - 1));
    my $pix = $cell->{pix};
    my $pi = 0;
		for (my $py = 0; $py < $self->{reso}; ++$py)
		{
			for (my $px = 0; $px < $self->{reso}; ++$px)
			{
				 # my ($r,$g,$b) = (, $pix->[$pi]->[1], $pix->[$pi]->[2]);
				 my $r = $pix->[$pi]->[0]*$alpha + 255*(1-$alpha);
				 my $g = $pix->[$pi]->[1]*$alpha + 255*(1-$alpha);
				 my $b = $pix->[$pi]->[2]*$alpha + 255*(1-$alpha);
				 # Draw pixel at right location
				 my $lval = sprintf("pixel[%d,%d]",
				 								$cell->{x}*$self->{reso}+$px,$cell->{y}*$self->{reso}+$py);
				 my $rval = sprintf("#%02x%02x%02x", $r,$g,$b);
         # print "$str\n";	
				 $heatmap->Set($lval => $rval);
				 $pi++;
			}
		}
		$n++;
  }
  # save heatmap
  print "Writing $oname ...\n";
	$heatmap->Write($oname);
  undef $heatmap;
}

sub sample_photos($)
{
  my ($self) = @_;
  my @images = ();

  die "No imageset provided\n" if !$self->{imageset};

	print "Sampling source images...\n" if $self->{verbose};
  my $idx = 0;
  while (1)
	{
			my $image = $self->{imageset}->get_image($idx, $self->{reso});
			last if !$image;

			# get dimensions
			my ($w, $h) = $image->Get('width', 'height');

			# kill bad choices... (overly uniform images)
			my $badImage = 0;

			if ($self->{noborders}) {
				my ($w1, $h1) = ($w-1,$h-1);
				my ($w2, $h2) = ($w/2,$h/2);
				my ($r1,$g1,$b1) = $image->GetPixels(x=>$w2,y=>0,width=>1,height=>1,normalize=>1);
				my ($r2,$g2,$b2) = $image->GetPixels(x=>$w2,y=>$h1,width=>1,height=>1,normalize=>1);
				my ($r3,$g3,$b3) = $image->GetPixels(x=>0,y=>$h2,width=>1,height=>1,normalize=>1);
				my ($r4,$g4,$b4) = $image->GetPixels(x=>$w1,y=>$h2,width=>1,height=>1,normalize=>1);
				my $d1 =  ($r2*255 - $r1*255)**2 +
									($g2*255 - $g1*255)**2 +
									($b2*255 - $b1*255)**2;
				my $d2 =  ($r4*255 - $r3*255)**2 +
									($g4*255 - $g3*255)**2 +
									($b4*255 - $b3*255)**2;
				if ($d1 <= 40 || $d2 <= 40 || $w/$h >= 2 || $h/$w >= 2) {
					$badImage = 1;
					print "." if $self->{verbose};
				}
			}

			if (!$badImage)
			{

				if ($w < $h) {
					$image->Crop(width=>$w, height=>$w, x=>0, y=>($h-$w)/2);
				}
				elsif ($w > $h) {
					$image->Crop(width=>$h, height=>$h, y=>0, x=>($w-$h)/2);
				}
				$image->Resize(width=>1,height=>1);
				my ($r,$g,$b) = $image->GetPixels(x=>0,y=>0,width=>1,height=>1,normalize=>1);
				my $l = 0.3086*$r + 0.6094*$g + 0.0820*$b; # Haeberli

				my $photo = {idx=>$idx, l=>$l};
				push @images, $photo;
			}
			undef $image;
			++$idx;
			print "$idx...\n" if $idx % 500 == 0 && $self->{verbose};
	}
	my $num_images = scalar(@images);
	print "Got $num_images images\n";
  $self->{images} = \@images;
}

sub setup_cells()
{
  my $self = shift;

	my @cells = ();

	my $baseimg = Image::Magick->new;
	my $err = $baseimg->Read($self->{basepic});
	warn "$err" if "$err";

	my ($w, $h) = $baseimg->Get('width', 'height');
	my $aspect = $h/$w;
	$self->{max_images} = scalar(@{$self->{images}})/4 if !$self->{max_images};
	$self->{max_images} = scalar(@{$self->{images}}) if $self->{images} && $self->{max_images} > scalar(@{$self->{images}});
	my $hcells = sqrt($self->{max_images} / $aspect);
	my $vcells = $hcells * $aspect;
	$self->{hcells} = int($hcells + .5);
	$self->{vcells} = int($vcells + .5);
	if ($self->{hcells} * $self->{vcells} > $self->{max_images})
	{
		# round down to fit...
		$self->{hcells} = int($hcells);
		$self->{vcells} = int($vcells);
	}

	# if $reso is too large for target image, reduce it...
	if ($self->{reso}*$self->{hcells} > $w)
	{
		$self->{reso} = int($w / $self->{hcells});
		$self->{reso} = 1 if ($self->{reso} < 1);
		print "Forcing Reso to $self->{reso} due to lack of resolution in target image\n";
		$self->{reso2} = $self->{reso}*$self->{reso};
	}

  print "Allocating cell data ($self->{hcells} x $self->{vcells}) x $self->{reso} ...\n" if $self->{verbose};
	
	my $baseimg2 = Image::Magick->new;
	$err = $baseimg2->Read($self->{basepic});
	warn "$err" if "$err";

	$baseimg->Resize(width=>$self->{hcells}, height=>$self->{vcells}, filter=>$self->{targetfilter}, blur=>$self->{targetblur});
	$baseimg2->Resize(width=>$self->{hcells}*$self->{reso}, height=>$self->{vcells}*$self->{reso}, filter=>$self->{targetfilter}, blur=>$self->{targetblur});

	my $i = 0;
	for (my $y = 0; $y < $self->{vcells}; ++$y)
	{
		for (my $x = 0; $x < $self->{hcells}; ++$x)
		{
			my ($r,$g,$b) = $baseimg->GetPixels(x=>$x,y=>$y,width=>1,height=>1,normalize=>1);
			my $l = 0.3086*$r + 0.6094*$g + 0.0820*$b; # Haeberli
			# these values currently unused
			# my ($h, $s, $v) = RGBtoHSV($r,$g,$b);

			my ($x0,$y0) = ($x*$self->{reso}, $y*$self->{reso});

			my @pix = ();

			for (my $py = 0; $py < $self->{reso}; ++$py)
			{
				for (my $px = 0; $px < $self->{reso}; ++$px)
				{
					my ($x1,$y1) = ($x0+$px, $y0+$py);
					my ($r1,$g1,$b1) = $baseimg2->GetPixels(x=>$x1,y=>$y1,width=>1,height=>1,normalize=>1);
					push @pix, [$r1, $g1, $b1];
				}
			}
			
			push @cells, {i=>$i, x=>$x, y=>$y, l=>$l, pix=>\@pix};
			$i++;
		}
	}
	undef $baseimg;
	undef $baseimg2;
	$self->{cells} = \@cells;
  print "Sorting cells...\n" if $self->{verbose};
  $self->{sortedcells} = [reverse sort { $self->Edginess($a) <=> $self->Edginess($b); } @{$self->{cells}}];
}

# Build index into luminence values in pre-sorted images
#

sub BuildLumIndex($)
{
  my $self = shift;

  $self->{images} = [sort {$a->{l} <=> $b->{l}} @{$self->{images}}];

	my @iIndex = ();
	my $lIdx = -1;
	my $n = 0;
  my $j = 0;
  my $num = scalar(@{$self->{images}});

  print "Sorting $num images for luminance\n" if $self->{verbose} > 1;

	for my $img (@{$self->{images}})
	{
		 if (int($img->{l}*255) != $lIdx)
		 {
			 $lIdx = int($img->{l}*255);
			 while ($n <= $lIdx)
			 {
				$iIndex[$n++] = $j;
			 }
		 }
		 $j++;
	}
	while ($n <= 255) {
		$iIndex[$n++] = $j;
	}
	printf "Lumindex has $n entries\n" if $self->{verbose} > 1;
	$self->{iIndex} = \@iIndex;
}

sub select_tiles()
{
  my ($self) = @_;

	$self->sample_photos() if !$self->{images};
  return if !$self->{images};
  $self->setup_cells() if !$self->{sortedcells};
  return if !$self->{sortedcells};

	# sort images by luminance and build histogram index
	# to speed up matching
	my $numImages = scalar( @{$self->{images}} );
	my $lastImageIdx = $numImages-1;
	print "Selecting from $numImages images...\n" if $self->{verbose};

	$self->BuildLumIndex();
	print "Lum index done..\n"  if $self->{verbose} > 1;	
  my $i = 0;
  my @fimages = ();
  
	my $stime = time();
	for my $cell (@{$self->{sortedcells}})
	{
    print "tile $i  cell $cell->{x} x $cell->{y} $cell->{l}  ", if $self->{verbose} > 1;
		# 	Find the closest match in @images
		my $cIdx;
		my $minDiff = 256*256*256;
		my $flop;
		my $lErr = $numImages > 10000? 1 : ($numImages > 5000? 5 : 10);  # 10 is an improvement over 1...
		my $gotOne = 0;
		
		while (!$gotOne)
		{
			# use iIndex to find a group of photos which have the approximate
			# overall desired luminance
			my $ii = int($cell->{l}*255);
			my $mini = $self->{iIndex}->[$ii - $lErr < 0? 0 : $ii - $lErr];
			my $maxi = $self->{iIndex}->[$ii + $lErr > 255? 255 : $ii + $lErr];
			if ($maxi - $mini < 20) 
			{
				$mini -= 10;
				$maxi += 10;
			}
			$mini = 0 if $mini < 0;
			$maxi = $lastImageIdx if $maxi > $lastImageIdx;

			print "$mini - $maxi\n" if $self->{verbose} > 1;
			# loop thru those cells and find the best fit
			for (my $j = $mini; $j <= $maxi; ++$j)
			{
			  my $image = $self->{images}->[$j];
			
				next if $image->{xx};

				$self->subsample_photo($image);	# subsample photo if we haven't looked at it yet

				my $diff = $self->CumDiff($image, $cell, $minDiff);
				if ($diff < $minDiff) {
					$minDiff = $diff;
					$cIdx = $j;
					$flop = 0;
					$gotOne = 1;
				}

				if ($self->{doflops}) 
				{
					$diff = $self->CumDiffFlop($image, $cell, $minDiff);
					if ($diff < $minDiff) {
						$minDiff = $diff;
						$cIdx = $j;
						$flop = 1;
						$gotOne = 1;
					}
				}
			}
			# if no match was find, widen the range and search again
			$lErr += 5;
		}
		# clone to a record in @fimages and set {i} to $cells[$i]->{i}
		my $cPhoto = clone($self->{images}->[$cIdx]);
		$cPhoto->{i} = $cell->{i};
		$cPhoto->{flop} = $flop;
		push @fimages,$cPhoto;

		# 	if no dupes, remove the record from @images
		$self->{images}->[$cIdx]->{xx} = 1 if !$self->{dupesOK};

		print "$i...\n" if ++$i % 100 == 0 && $self->{verbose};
	}
	printf "Done: elapsed = %d seconds\n", time() - $stime if $self->{verbose};
	$self->{finalimages} = \@fimages;

	undef $self->{images};
	undef $self->{iIndex};
	undef $self->{sortedcells};
}


sub generate_mosaic()
{
  my ($self, $iopts) = @_;

  $self->select_tiles() if !$self->{finalimages};
  return if !$self->{finalimages};

	my $rootname = $self->{basepic};
	$rootname =~ s/\.jpg//;
	my %opts = (filename => "pmosaic_${rootname}_$self->{reso}x$self->{reso}_$self->{hcells}_x_$self->{vcells}.jpg", 
	            quality=>90, 
	            width=>$self->{cellsize} * $self->{hcells}
	            );

  for my $k (keys %{$iopts})
  {
    $opts{$k} = $iopts->{$k};
  }
  
  my $cellsize = int($opts{width} / $self->{hcells} + .5);

	my ($width, $height) = ($cellsize*$self->{hcells}, 
	                        $cellsize*$self->{vcells});
  print "Image Dimensions will be $width x $height\n" if $self->{verbose};

	my $mosaic = Image::Magick->new;
	$mosaic->Set(size=>"${width}x${height}"); # worked okay at 10000x10000
	my $err = $mosaic->Read('xc:black');
	warn "$err" if "$err";

	my $i = 0;
	foreach my $imgdat (@{$self->{finalimages}})
	{
	  my $cell = $self->{cells}->[$imgdat->{i}];
		my ($x, $y) = ($cell->{x}, $cell->{y});
		
		my $img = $self->{imageset}->get_image($imgdat->{idx}, $cellsize);

		my ($w, $h) = $img->Get('width', 'height');
		if ($w < $h) {
			$img->Crop(width=>$w, height=>$w, x=>0, y=>($h-$w)/2);
		}
		elsif ($w > $h) {
			$img->Crop(width=>$h, height=>$h, y=>0, x=>($w-$h)/2);
		}
		$img->Resize(width=>$cellsize, height=>$cellsize, filter=>'Sinc', blur=>0.4);

		$img->Flop() if ($imgdat->{flop});

		$err = $mosaic->Composite(image=>$img, x=>$x*$cellsize, y=>$y*$cellsize);
		warn "$err" if "$err";

		undef $img;
		print "$i...\n" if (++$i % 500 == 0);
	}

	print "Saving JPEG $opts{filename}...\n" if $self->{verbose};

	$mosaic->Set(quality=>$opts{quality});
	$mosaic->Write($opts{filename});
	undef $mosaic;
}

1;